/***********************************************************************
 * Time-stamp: <Thursday, 19.07.2001 at 19:13:50; edited by bauke>
 * 
 * Tina's random number generators
 *
 * Copyright by Heiko Bauke, Magdeburg 2001 
 *
 * heiko.bauke@student.uni-magdeburg.de
 *
 * This is free software. You may use and distribute this software under
 * the terms of the GNU GENERAL PUBLIC LICENSE Version 2. See COPYING
 * for details.
 *
 ***********************************************************************/

#ifndef TRNG_H
#define TRNG_H

#include <vector>

namespace TRNG {

  enum RNG_type {RNG_t, generic_MLCG_t, ParkMiller_t, RAND_t, LCG64_t,
		 LMC2_t, LMC3_t, LMC4_t, CLCG2_t, CLCG3_t, CLCG4_t, 
		 EINV_t, EINVLCG64_t, user1_t, user2_t, user3_t}; 
  
  typedef struct {
    double x1, x2;
  } vector2d;
  
  typedef struct {
    double x1, x2, x3;
  } vector3d;
  
  typedef struct {
    double x1, x2, x3, x4;
  } vector4d;

  const char * version(void);
  class error {
  public:
    const char * message;
    error(const char * m) : message(m) {}
  };
  long modulo_invers(long, long);
  void gauss(std::vector<long> &a, std::vector<long> &b, long);
  void matrix_mult(std::vector<long> &, std::vector<long> &, 
		   std::vector<long> &, long);
  void matrix_vec_mult(std::vector<long> &, std::vector<long> &, 
		   std::vector<long> &, long);
  long long scalar_prod(std::vector<long long> &, 
			std::vector<long long> &);
  
  double ln_Gamma(double);
  double Gamma(double);
  double GammaP(double, double);
  double GammaQ(double, double);
  double incomp_Gamma(double, double);
  double comp_incomp_Gamma(double, double);
  double Gamma_ser(double, double);
  double Gamma_cf(double, double);
  double Chisquare(std::vector<double> &, std::vector<double> &);
  double Chisquare_prob(double, long);
  double errf(double);
  double Stirling_num2(long, long);
  double Student_t(double, long, bool=true);

  class RNG {
    // generic class for random number generators
  protected:
    long max_val;
    long max_val2;
  public:
    class error {
    public:
      const char * message;
      error(const char * m) : message(m) {}
    };
    static const RNG_type type=RNG_t;
    virtual const char * name(void);
    virtual void reset(void);
    virtual void seed(long);
    virtual long rand(void);
    bool boolean(void);
    bool boolean(double);
    double uniform(void);
    double uniformco(void);
    double uniformcc(void);
    double uniformoo(void);
    double uniformoc(void);

    double exp_dist(double=1.0);
    double normal_dist(double=1.0, double=0.0);
    double Gamma_dist(double);
    double Beta_dist(double, double);
    double chi_square_dist(double);
    double Student_t_dist(double);
    long binomial_dist(long, double=0.5);
    long poisson_dist(double=1.0);
    double rejection(double (*)(double), double, double, double);
    vector2d spherical2d(void);
    vector3d spherical3d(void);
    vector4d spherical4d(void);
    
    long max(void);
    virtual RNG *copy(void);
    virtual void split(long, long);
    virtual void jump(long);
    virtual void save_status(std::vector<long> &);
    virtual void load_status(std::vector<long> &);
    RNG();
  };
  
  class generic_MLCG : public RNG {
    // generic class for multiplicative linear congruential 
    // random number generators
  private:
    long a;
    long a_save;
    long r;
    long r_save;
    long modulus;
    long modulus_save;
    void backward(void);
  public:
    static const RNG_type type=generic_MLCG_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    generic_MLCG(long, long, long);
  };

  class ParkMiller : public RNG {
    // modified random number generator by Park and Miller
    // see Communication of the ACM, Oct. 1988, Vol. 31, Nr. 10, 
    // pp. 1192-1201
  private:
    generic_MLCG R;
  public:
    static const RNG_type type=ParkMiller_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    ParkMiller();
  };
  
  class RAND : public RNG {
  private:
    unsigned long a, b, r;
    void backward(void);
  public:
    static const RNG_type type=RAND_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    RAND();
  };

  class LCG64 : public RNG {
    // linear congruential random number generator 
    // modulo 2^64
  private:
    unsigned long long a;
    unsigned long long b;
    unsigned long long r;
    void backward(void);
  public:
    static const RNG_type type=LCG64_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    LCG64();
  };

  class LMC2 : public RNG {
    // multiplicative linear congruential random number generator
    // with 2 multipliers
  private:
    std::vector<long> a;
    std::vector<long> r;
    long modulus;
    void backward(void);
  public:
    static const RNG_type type=LMC2_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    LMC2();
    virtual ~LMC2() {}
  };

  class LMC3 : public RNG {
    // multiplicative linear congruential random number generator
    // with 3 multipliers
  private:
    std::vector<long> a;
    std::vector<long> r;
    long modulus;
    void backward(void);
  public:
    static const RNG_type type=LMC3_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    LMC3();
    virtual ~LMC3() {}
  };

  class LMC4 : public RNG {
    // multiplicative linear congruential random number generator
    // with 4 multipliers
  private:
    std::vector<long> a;
    std::vector<long> r;
    long modulus;
    void backward(void);
  public:
    static const RNG_type type=LMC4_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    LMC4();
    virtual ~LMC4() {}
  };

  class CLCG2 : public RNG {
    // combined linear congruential random number generator 
    // with two generators
  private:
    generic_MLCG R1;
    generic_MLCG R2;
    static const long a1= 376555083l, m1=2147482951l;
    static const long a2=1028879659l, m2=2147482949l;
    long modulus;
  public:
    static const RNG_type type=CLCG2_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    CLCG2();
  };

  class CLCG3 : public RNG {
    // combined linear congruential random number generator 
    // with three generators
  private:
    generic_MLCG R1;
    generic_MLCG R2;
    generic_MLCG R3;
    static const long a1= 376555083l, m1=2147482951l;
    static const long a2=1028879659l, m2=2147482949l;
    static const long a3= 225802979l, m3=2147482943l;
    long modulus;
  public:
    static const RNG_type type=CLCG3_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    CLCG3();
  };

  class CLCG4 : public RNG {
    // combined linear congruential random number generator 
    // with four generators
  private:
    generic_MLCG R1;
    generic_MLCG R2;
    generic_MLCG R3;
    generic_MLCG R4;
    static const long a1= 376555083l, m1=2147482951l;
    static const long a2=1028879659l, m2=2147482949l;
    static const long a3= 225802979l, m3=2147482943l;
    static const long a4=2028073966l, m4=2147482859l;
    long modulus;
  public:
    static const RNG_type type=CLCG4_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    CLCG4();
  };
  
  class EINV : public RNG {
    // explicit inversive congruential generator
    // Eichenauer-Hermann, J. "Statistical independence of a new class 
    // of inversive congruential pseudorandom numbers"
    // Math. Comp. 60:375-384, 1993
  private:
    long a, i, di, modulus;
    long modulo_invers(long);
  public:
    static const RNG_type type=EINV_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    EINV();
  };

  class EINVLCG64 : public RNG {
    // combined generator
    // explicit inversive congruential generator and 
    // linear congruential random number generator modulo 2^64 
  private:
    LCG64 R1;
    EINV R2;
  public:
    static const RNG_type type=EINVLCG64_t;
    const char * name(void);
    void reset(void);
    void seed(long);
    long rand(void);
    RNG *copy(void);
    void split(long, long);
    void jump(long);
    void save_status(std::vector<long> &);
    void load_status(std::vector<long> &);
    EINVLCG64();
  };

}

#endif
